/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001-2003,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <global.h>
#include <unistd.h>

regionvec regions;

region *
region::default_region ()
{
  for_all_regions (rgn)
    if (rgn->fallback)
      return rgn;

  return regions [0];
}

/* 
 * Pass a char array, returns a pointer to the region of the same name.
 * if it can't find a region of the same name it returns the first region
 * with the 'fallback' property set.
 * if it can't find a matching name /or/ a fallback region it logs an info message 
 * message and returns NULL
 * used by the map parsing code. 
 */
region *
region::find (shstr_cmp name)
{
  for_all_regions (rgn)
    if (rgn->name == name)
      return rgn;

  LOG (llevError, "region called %s requested, but not found, using fallback.\n", &name);

  return default_region ();
}

void
region::do_destroy ()
{
  regions.erase (this);

  attachable::do_destroy ();

  refcnt_dec ();
}

/* 
 * returns 1 if the player is in the region reg, or a child region thereof
 * otherwise returns 0
 * if passed a NULL region returns -1
 */
static int
region_is_child_of_region (const region * child, const region * r)
{
  if (!r)
    return -1;

  if (!child)
    return 0;

  if (child->name == r->name)
    return 1;

  if (child->parent)
    return region_is_child_of_region (child->parent, r);

  return 0;
}

/** Returns an object which is an exit through which the player represented by op should be
 *  sent in order to be imprisoned. If there is no suitable place to which an exit can be 
 *  constructed, then NULL will be returned. The caller is responsible for freeing the object
 *  created by this function.
 */
object *
get_jail_exit (object *op)
{
  if (op->type != PLAYER)
    {
      LOG (llevError, "region.c: get_jail_exit called against non-player object.\n");
      return NULL;
    }

  region *reg = op->region ();
  while (reg)
    {
      if (reg->jailmap)
        {
          object *exit = object::create ();
          EXIT_PATH (exit) = reg->jailmap;
          /* damned exits reset savebed and remove teleports, so the prisoner can't escape */
          SET_FLAG (exit, FLAG_DAMNED);
          EXIT_X (exit) = reg->jailx;
          EXIT_Y (exit) = reg->jaily;
          return exit;
        }
      else
        reg = reg->parent;
    }

  LOG (llevError, "No suitable jailmap for region %s was found.\n", &reg->name);

  return 0;
}

region *
region::read (object_thawer &f)
{
  assert (f.kw == KW_region);

  region *rgn = new region;
  rgn->refcnt_inc ();

  f.get (rgn->name);
  f.next ();

  for (;;)
    {
      switch (f.kw)
        {
          case KW_parent:
            rgn->parent = region::find (f.get_str ());
            break;

          case KW_msg:        f.get_ml (KW_endmsg, rgn->msg); break;
          case KW_longname:   f.get (rgn->longname); break;
          case KW_jail_map:   f.get (rgn->jailmap);  break;
          case KW_jail_x:     f.get (rgn->jailx);    break;
          case KW_jail_y:     f.get (rgn->jaily);    break;
          case KW_portal_map: f.get (rgn->portalmap);break;
          case KW_portal_x:   f.get (rgn->portalx);  break;
          case KW_portal_y:   f.get (rgn->portaly);  break;
          case KW_fallback:   f.get (rgn->fallback); break;
          case KW_chance:     f.get (rgn->treasure_density); break;

          case KW_randomitems:
            rgn->treasure = treasurelist::get (f.get_str ());
            break;

          case KW_end:
            f.next ();

            // cannot use find as that will request the default region
            for_all_regions (old)
              if (old->name == rgn->name)
                {
                  old->destroy ();
                  break;
                }

            // just append
            regions.push_back (rgn);
            return rgn;

          case KW_ERROR:
            rgn->set_key_text (f.kw_str, f.value);
            //fprintf (stderr, "region addkv(%s,%s)\n", f.kw_str, f.value);//D
            break;

          default:
            if (!f.parse_error ("region", rgn->name))
              {
                rgn->destroy ();
                return 0;
              }
            break;
        }

      f.next ();
    }
}

/*
 * First initialises the archtype hash-table (init_archetable()).
 * Reads and parses the archetype file (with the first and second-pass
 * functions).
 */
void
init_regions (void)
{
  if (!regions.size ())
    {
      // make sure one region is always available
      region *rgn = new region;
      rgn->name = "<builtin>";
      rgn->longname = "Built-in Region";
      regions.push_back (rgn);
    }
}

